//
// I2CLiquidCrystal.h
//
// Class to control BME280
// (See. BST-BME280_DS001-10.pdf)
// 
// Written by Nobuki HIRAMINE
// http://www.hiramine.com
// Last modified : 2017/07/30
//

#include <SPI.h>

#define FORCEDMODE	(0b0001)

// オプションパラメータ
struct OptionParameters
{
	uint8_t		ui8TemperatureOversampling;
	uint8_t		ui8HumidityOversampling;
	uint8_t		ui8PressureOversampling;
	uint8_t		ui8Mode;
	uint8_t		ui8StandbyTime;
	uint8_t		ui8Filter;
	uint8_t		ui8SPI3WireEnabled;
};

// トリミングパラメータ
struct TrimmingParameters
{
	// 温度関連のトリミングパラメータ
	uint16_t	ui16DigT1;
	 int16_t	 i16DigT2;
	 int16_t	 i16DigT3;
	// 気圧関連のトリミングパラメータ
	uint16_t	ui16DigP1;
	 int16_t	 i16DigP2;
	 int16_t	 i16DigP3;
	 int16_t	 i16DigP4;
	 int16_t	 i16DigP5;
	 int16_t	 i16DigP6;
	 int16_t	 i16DigP7;
	 int16_t	 i16DigP8;
	 int16_t	 i16DigP9;
	// 湿度関連のトリミングパラメータ
	uint8_t		ui8DigH1;
	 int16_t	 i16DigH2;
	uint8_t		ui8DigH3;
	 int16_t	 i16DigH4;
	 int16_t	 i16DigH5;
	 int8_t		 i8DigH6;
};

class BME280
{
private:
	int m_pinCSB;
	OptionParameters	m_optionparams;
	TrimmingParameters 	m_trimparams;
	int m_iTimeMeasurementMilliSec;
	
public:
	// コンストラクタ
	BME280( int     pinCSB,
			uint8_t ui8TemperatureOversampling,
			uint8_t ui8HumidityOversampling,
			uint8_t ui8PressureOversampling,
			uint8_t ui8Mode,
			uint8_t ui8StandbyTime,
			uint8_t ui8Filter,
			uint8_t ui8SPI3WireEnabled )
	{
		m_pinCSB = pinCSB;
		m_optionparams.ui8TemperatureOversampling	= ui8TemperatureOversampling;
		m_optionparams.ui8HumidityOversampling		= ui8HumidityOversampling;
		m_optionparams.ui8PressureOversampling		= ui8PressureOversampling;
		m_optionparams.ui8Mode						= ui8Mode;
		m_optionparams.ui8StandbyTime				= ui8StandbyTime;
		m_optionparams.ui8Filter					= ui8Filter;
		m_optionparams.ui8SPI3WireEnabled			= ui8SPI3WireEnabled;
		
		// 測定に要する時間 (See. P51, 9.1 Measurement time)
		// TimeMeasurementMax = 1.25 + (2.3 * TempOversampling) + (2.3 * PressureOversampling + 0.575) + (2.3 * HumidityOversampling + 0.575)
		double dTimeMeasurementMax = 1.25
		                           + (2.3 * getOversamplingCount(m_optionparams.ui8TemperatureOversampling))
		                           + (2.3 * getOversamplingCount(m_optionparams.ui8PressureOversampling) + 0.575)
		                           + (2.3 * getOversamplingCount(m_optionparams.ui8HumidityOversampling) + 0.575);
		m_iTimeMeasurementMilliSec = (int)dTimeMeasurementMax + 1;	// 「+1」で簡易切り上げ演算を行う。
	}

	// 開始
	void begin()
	{
		// スレーブセレクトピンの設定
		pinMode( m_pinCSB, OUTPUT );

		// SPIのセットアップ
		SPI.begin();
		SPI.setBitOrder( MSBFIRST );	// most-significant bit first

		setOptions();
		
		readTrimmingParameters();
	}

	// 温度、湿度、気圧の取得
	void getData(	double& rdTemperature,
					double& rdHumidity,
					double& rdPressure )
	{
		if( 0b0011 != m_optionparams.ui8Mode )
		{	// インスタンス時に設定したモードがノーマルモード以外の場合は、
			// スリープ状態にあるので、フォースドモードにモード変更することで、１回測定を実施する。
			// （インスタンス時に設定したモードがノーマルモードの場合は、
			//   ノーマル状態にあるので、「測定に要する時間＋スタンバイタイム」間隔で測定が実施されているので、測定に関するアクションは不要）

			// 測定制御についての設定 (See. P26, 5.4.5 Register 0xF4 “ctrl_meas”)
			uint8_t ui8MeasurementControl = (m_optionparams.ui8TemperatureOversampling << 5) | (m_optionparams.ui8PressureOversampling << 2) | FORCEDMODE;
			writeRegister( 0xF4, ui8MeasurementControl );	// レジスタへの書き込み
			// 測定完了を待つ。
			delay(m_iTimeMeasurementMilliSec);
		}
		
		// ADC値の読み込み
		uint32_t ui32Temperature_adc, ui32Humidity_adc, ui32Pressure_adc;
		getAdcData( ui32Temperature_adc, ui32Humidity_adc, ui32Pressure_adc );

		// ADC値の補正
		int32_t i32Temperature_fine = 0;
		int32_t i32Temperature_comp = BME280_compensate_T_int32( ui32Temperature_adc, i32Temperature_fine );
		int32_t i32Humidity_comp    = bme280_compensate_H_int32( ui32Humidity_adc, i32Temperature_fine );
		int32_t i32Pressure_comp    = BME280_compensate_P_int64( ui32Pressure_adc, i32Temperature_fine );

		// 人間にとっての値に変換し、出力変数へ代入
		// (See. P23, 4.2.3 Compensation formulas)
		rdTemperature = (double)i32Temperature_comp * 0.01;			// [DegC]
		rdHumidity    = (double)i32Humidity_comp / 1024.0;			// [%RH]
		rdPressure    = (double)i32Pressure_comp / 256.0 * 0.01;	// [hPa]
	}
	
private:
	// BME280のセットアップ
	// (See. P25, 5.4 Register description)
	void setOptions()
	{
		// 湿度制御についての設定 (See. P25, 5.4.3 Register 0xF2 “ctrl_hum”)
		uint8_t ui8HumidityControl = (m_optionparams.ui8HumidityOversampling);
		writeRegister( 0xF2, ui8HumidityControl );	// レジスタへの書き込み
		// 測定制御についての設定 (See. P26, 5.4.5 Register 0xF4 “ctrl_meas”)
		uint8_t ui8MeasurementControl = (m_optionparams.ui8TemperatureOversampling << 5) | (m_optionparams.ui8PressureOversampling << 2) | (m_optionparams.ui8Mode);
		writeRegister( 0xF4, ui8MeasurementControl );	// レジスタへの書き込み
		// 構成についての設定 (See. P27, 5.4.6 Register 0xF5 “config”)
		uint8_t ui8Config = (m_optionparams.ui8StandbyTime << 5) | (m_optionparams.ui8Filter << 2) | (m_optionparams.ui8SPI3WireEnabled);
		writeRegister( 0xF5, ui8Config );	// レジスタへの書き込み
	}

	// トリミングパラメータの読み込み
	// (See. P22, 4.2.2 Trimming parameter readout)
	void readTrimmingParameters()
	{
		// 温度関連のトリミングパラメータ
		m_trimparams.ui16DigT1 =            readRegister(0x88) | (readRegister(0x89) << 8);
		m_trimparams.i16DigT2  = (int16_t)( readRegister(0x8A) | (readRegister(0x8B) << 8) );
		m_trimparams.i16DigT3  = (int16_t)( readRegister(0x8C) | (readRegister(0x8D) << 8) );
		// 気圧関連のトリミングパラメータ
		m_trimparams.ui16DigP1 =            readRegister(0x8E) | (readRegister(0x8F) << 8);
		m_trimparams.i16DigP2  = (int16_t)( readRegister(0x90) | (readRegister(0x91) << 8) );
		m_trimparams.i16DigP3  = (int16_t)( readRegister(0x92) | (readRegister(0x93) << 8) );
		m_trimparams.i16DigP4  = (int16_t)( readRegister(0x94) | (readRegister(0x95) << 8) );
		m_trimparams.i16DigP5  = (int16_t)( readRegister(0x96) | (readRegister(0x97) << 8) );
		m_trimparams.i16DigP6  = (int16_t)( readRegister(0x98) | (readRegister(0x99) << 8) );
		m_trimparams.i16DigP7  = (int16_t)( readRegister(0x9A) | (readRegister(0x9B) << 8) );
		m_trimparams.i16DigP8  = (int16_t)( readRegister(0x9C) | (readRegister(0x9D) << 8) );
		m_trimparams.i16DigP9  = (int16_t)( readRegister(0x9E) | (readRegister(0x9F) << 8) );
		// 湿度関連のトリミングパラメータ
		m_trimparams.ui8DigH1  =            readRegister(0xA1);
		m_trimparams.i16DigH2  = (int16_t)( readRegister(0xE1) | (readRegister(0xE2) << 8) );
		m_trimparams.ui8DigH3  =            readRegister(0xE3);
		uint8_t e5 = readRegister(0xE5);
		m_trimparams.i16DigH4  = (int16_t)( (readRegister(0xE4) << 4) | (e5 & 0x0F) );
		m_trimparams.i16DigH5  = (int16_t)( (readRegister(0xE6) << 4) | (e5 >> 4) );
		m_trimparams.i8DigH6   =    (int8_t)readRegister(0xE7);
	}

	// ADC値の取得（温度、湿度、気圧）
	// (See. P28, 5.4.7 Register 0xF7…0xF9 “press” (_msb, _lsb, _xlsb))
	void getAdcData( uint32_t& rui32Temperature_adc, uint32_t& rui32Humidity_adc, uint32_t& rui32Pressure_adc )
	{
		// 気圧値の読み込み
		uint32_t ui32Pressure_msb	= readRegister(0xF7);
		uint32_t ui32Pressure_lsb	= readRegister(0xF8);
		uint32_t ui32Pressure_xlsb	= readRegister(0xF9);
		// 温度値の読み込み
		uint32_t ui32Temperature_msb	= readRegister(0xFA);
		uint32_t ui32Temperature_lsb	= readRegister(0xFB);
		uint32_t ui32Temperature_xlsb	= readRegister(0xFC);
		// 湿度値の読み込み
		uint32_t ui32Humidity_msb	= readRegister(0xFD);
		uint32_t ui32Humidity_lsb	= readRegister(0xFE);

		// ADC値の計算
		rui32Pressure_adc = (ui32Pressure_msb << 12) | (ui32Pressure_lsb << 4) | (ui32Pressure_xlsb >> 4);
		rui32Temperature_adc = (ui32Temperature_msb << 12) | (ui32Temperature_lsb << 4) | (ui32Temperature_xlsb >> 4);
		rui32Humidity_adc = (ui32Humidity_msb << 8) | (ui32Humidity_lsb);
	}

	// 温度ADC値の補正
	// (See. P23, 4.2.3 Compensation formulas)
	// Returns temperature in DegC, resolution is 0.01 DegC. Output value of “5123” equals 51.23 DegC.
	// t_fine carries fine temperature as global value
	int32_t BME280_compensate_T_int32(int32_t adc_T, int32_t& ri32Temperature_fine )
	{
		int32_t var1, var2, T;
		var1 = ((((adc_T>>3) - ((int32_t)m_trimparams.ui16DigT1<<1))) * ((int32_t)m_trimparams.i16DigT2)) >> 11;
		var2 = (((((adc_T>>4) - ((int32_t)m_trimparams.ui16DigT1)) * ((adc_T>>4) - ((int32_t)m_trimparams.ui16DigT1))) >> 12) * ((int32_t)m_trimparams.i16DigT3)) >> 14;
		ri32Temperature_fine = var1 + var2;
		T = (ri32Temperature_fine * 5 + 128) >> 8;
		return T;
	}

	// 気圧ADC値の補正
	// (See. P23, 4.2.3 Compensation formulas)
	// Returns pressure in Pa as unsigned 32 bit integer in Q24.8 format (24 integer bits and 8 fractional bits).
	// Output value of “24674867” represents 24674867/256 = 96386.2 Pa = 963.862 hPa
	uint32_t BME280_compensate_P_int64(int32_t adc_P, int32_t i32Temperature_fine )
	{
		int64_t var1, var2, p;
		var1 = ((int64_t)i32Temperature_fine) - 128000;
		var2 = var1 * var1 * (int64_t)m_trimparams.i16DigP6;
		var2 = var2 + ((var1*(int64_t)m_trimparams.i16DigP5)<<17);
		var2 = var2 + (((int64_t)m_trimparams.i16DigP4)<<35);
		var1 = ((var1 * var1 * (int64_t)m_trimparams.i16DigP3)>>8) + ((var1 * (int64_t)m_trimparams.i16DigP2)<<12);
		var1 = (((((int64_t)1)<<47)+var1))*((int64_t)m_trimparams.ui16DigP1)>>33;
		if (var1 == 0)
		{
			return 0; // avoid exception caused by division by zero
		}
		p = 1048576-adc_P;
		p = (((p<<31)-var2)*3125)/var1;
		var1 = (((int64_t)m_trimparams.i16DigP9) * (p>>13) * (p>>13)) >> 25;
		var2 = (((int64_t)m_trimparams.i16DigP8) * p) >> 19;
		p = ((p + var1 + var2) >> 8) + (((int64_t)m_trimparams.i16DigP7)<<4);
		return (uint32_t)p;
	}

	// 湿度ADC値の補正
	// (See. P23, 4.2.3 Compensation formulas)
	// Returns humidity in %RH as unsigned 32 bit integer in Q22.10 format (22 integer and 10 fractional bits).
	// Output value of “47445” represents 47445/1024 = 46.333 %RH
	uint32_t bme280_compensate_H_int32(int32_t adc_H, int32_t i32Temperature_fine )
	{
		int32_t v_x1_u32r;
		v_x1_u32r = (i32Temperature_fine - ((int32_t)76800));
		v_x1_u32r = (((((adc_H << 14) - (((int32_t)m_trimparams.i16DigH4) << 20) - (((int32_t)m_trimparams.i16DigH5) * v_x1_u32r)) +
		((int32_t)16384)) >> 15) * (((((((v_x1_u32r * ((int32_t)m_trimparams.i8DigH6)) >> 10) * (((v_x1_u32r * 
		((int32_t)m_trimparams.ui8DigH3)) >> 11) + ((int32_t)32768))) >> 10) + ((int32_t)2097152)) *
		((int32_t)m_trimparams.i16DigH2) + 8192) >> 14));
		v_x1_u32r = (v_x1_u32r - (((((v_x1_u32r >> 15) * (v_x1_u32r >> 15)) >> 7) * ((int32_t)m_trimparams.ui8DigH1)) >> 4));
		v_x1_u32r = (v_x1_u32r < 0 ? 0 : v_x1_u32r);
		v_x1_u32r = (v_x1_u32r > 419430400 ? 419430400 : v_x1_u32r);
		return (uint32_t)(v_x1_u32r>>12);
	}

	// レジスターへの書き込み
	// (See. P33, 6.3.1 SPI write)
	void writeRegister( uint8_t ui8Address, uint8_t ui8Value )
	{
		digitalWrite( m_pinCSB, LOW );
		SPI.transfer( ui8Address & 0b01111111 );
		SPI.transfer( ui8Value );
		digitalWrite( m_pinCSB, HIGH );
	}

	// レジスターからの読み込み（8ビット値）
	// (See. P33, 6.3.2 SPI read)
	uint8_t readRegister( uint8_t ui8Address )
	{
		digitalWrite( m_pinCSB, LOW );
		SPI.transfer( ui8Address | 0b10000000 );
		uint8_t ui8Value = SPI.transfer( 0b00000000 );
		digitalWrite( m_pinCSB, HIGH );

		return ui8Value;
	}
	
	// オーバーサンプリングのビットフラグ値から回数を求める
	// (See. P26, 5.4.5 Register 0xF4 “ctrl_meas”)
	uint8_t getOversamplingCount( uint8_t bitflag )
	{
		switch( bitflag )
		{
		case 0b000:
			return 0;
		case 0b001:
			return 1;
		case 0b010:
			return 2;
		case 0b011:
			return 4;
		case 0b100:
			return 8;
		}
		return 16;
	}
};
